use pnet::{
    datalink::{self, Channel, ChannelType, Config},
    packet::{ip::IpNextHeaderProtocol, ipv4::MutableIpv4Packet, Packet},
    util,
};
use std::{env, error::Error};

fn main() -> Result<(), Box<dyn Error>> {
    // IPを表す識別子(EtherType)
    const INTERNET_PROTOCOL: u16 = 0x0800;
    // テスト用のプロトコルを表すプロトコル番号
    const TESTING: u8 = 0xFE;
    let args: Vec<String> = env::args().collect();
    // 第1引数はネットワークインターフェース名
    let if_name = &args[1];
    // 第2引数は宛先のIPアドレス
    let destination = args[2].parse()?;

    // ネットワークインターフェースの取得
    let iface = datalink::interfaces()
        .into_iter()
        .find(|i| i.name == *if_name)
        .ok_or("failed to get interface")?;
    let config = Default::default();
    // ソケットの生成
    let (mut sender, _receiver) = match datalink::channel(
        &iface,
        // channnel_type以外のフィールドはデフォルトの値を利用
        Config {
            channel_type: ChannelType::Layer3(INTERNET_PROTOCOL),
            ..config
        },
    ) {
        Ok(Channel::Ethernet(sender, receiver)) => (sender, receiver),
        _ => return Err("failed to open channel".into()),
    };

    // パケットのもとになるバッファ
    let mut ip_buf = [0; 1500];
    let mut packet = MutableIpv4Packet::new(&mut ip_buf).ok_or("failed to create a packet")?;

    // IPのバージョン番号。IPv4なので4
    packet.set_version(4);
    // ヘッダ長。単位は4オクテット(※)（32ビット）
    // 今回のヘッダは20オクテットなので5を指定
    packet.set_header_length(5);
    // IPパケットのペイロード含めた全長。単位はオクテット
    packet.set_total_length(1500);
    // パケットのID。フラグメンテーション（後述）に利用する。今回は適当
    packet.set_identification(0xabcd);
    // Time to Live. ルータを越えるたびに1減っていく
    // パケットの無限ループを防ぐためのもの
    packet.set_ttl(64);
    // IPの上位レイヤのプロトコル
    // 今回は考えないのでテスト用に予約された値（0xFE）を用いている
    packet.set_next_level_protocol(IpNextHeaderProtocol::new(TESTING));
    // 送信元IPアドレス
    packet.set_source("127.0.0.1".parse()?);
    // 宛先IPアドレス
    packet.set_destination(destination);
    // チェックサム。パケットの破損を検出するためのもの
    packet.set_checksum(util::checksum(&packet.packet(), 5));
    // send_toはOption<Result<(), Error>>を返すので??となっている
    sender
        .send_to(packet.packet(), None)
        .ok_or("failed to send")??;
    Ok(())
}
